import subprocess
import json
import aaf2
import sys, os
import codecs
import logging
import logging.handlers
import shutil
from tempfile import gettempdir, mkdtemp
from uuid import uuid4

def meta(rate):
    duration_ts = {'23.976':96096, '24':96000, '25':92160, '30':96000}
    samples_per_frame = {'23.976':2002, '24':2000, '25':1920, '30':1600}
    filename = {"23.976":23976, "24":24, "25":25, "30":24}
    info = {
            'streams':[
                {"sample_rate":48000, "channels":1,
                'duration_ts':duration_ts[rate]}
                ], 
            'format':{"filename": "BEEPS_{}.wav".format(filename[rate]), "tags":{"time_reference": "0"}},
            'samples_per_frame':samples_per_frame[rate]
            }
    return info

def createLoggerWithPath2(path):

    LOGGER_NAME = 'APP'

    LOGFILENAME = path
    formatter = logging.Formatter(fmt='%(asctime)s %(funcName)s %(message)s', 
        datefmt="%Y-%m-%d %H:%M:%S")
    handler = logging.handlers.RotatingFileHandler(LOGFILENAME, maxBytes=1024*1024*2, backupCount=3)
    handler.setFormatter(formatter)

    if LOGGER_NAME in [name for name in logging.root.manager.loggerDict]:
        return logging.getLogger(LOGGER_NAME)

    my_logger = logging.getLogger(LOGGER_NAME)
    my_logger.setLevel(logging.DEBUG)
    for h in my_logger.handlers:
        my_logger.removeHandler(h) 
    my_logger.addHandler(handler)
    my_logger.addHandler(logging.StreamHandler())
    return my_logger

class AAFWriter(object):
    """
    The inputFile is a txt file of json cue data
    the destinationFile is an aff file path
    """
    def __init__(self, inputFile, destinationFile, context):  
        """
        Top level obkects include:
        tracks,
        video_rate, 
        referenced,
        start in frames
        assume 24 frame
        """  
        super(AAFWriter, self).__init__()
        self.inputFile = inputFile
        self.destinationFile = destinationFile
        self.jsonData = None
        self.errorCode = 0
        self.error = None

        self.jsonData = self._processJsonData(self.inputFile)
        if self.errorCode != 0:
            logger.debug(self.errorCode)
            return

        # print json.dumps(self.jsonData, indent=4, sort_keys=True)

        try:
            self.video_rate = str(self.jsonData['video_rate'])
            logger.debug(self.video_rate)          
            if self.video_rate not in ['23.976', '24', '25', '30']:
                self.errorCode = -1
                self.error = 'Unsupported Video Rate'      
        except Exception as e:
            self.errorCode = -1
            self.error = 'No Video Rate Key'
            return

        try:
            self.tracks = self.jsonData['tracks']
            logger.debug(self.tracks)
        except Exception as e:
            self.errorCode = -1
            self.error = 'No tracks Key'
            return            

        try:
            self.shouldLink = self.jsonData.get('referenced', False)
            logger.debug(self.shouldLink)
        except Exception as e:
            self.errorCode = -1
            self.error = 'No referenced Key'
            return            

        try:
            self.startInFrames = self.jsonData.get('start', 0)
            logger.debug(self.startInFrames)
        except Exception as e:
            self.errorCode = -1
            self.error = 'No start Key'
            return        

        logger.debug(context)

        if context == 'beeps':
            self._writeAAFBeeps()
        elif context == 'clips':
            self._writeAAFClips()
        elif context == 'tenbyten':
            logger.debug('context: 10 BY 10')
            self._writeAAFTenByTen()


    def __repr__(self):
        return "<AAFWriter> {} {}".format(self.errorCode, self.error)

    def _processJsonData(self, filePath):

        try:
            logger.debug('reading json data 2')
            with open(filePath, 'r') as f:
                return json.loads(f.read())
        except Exception as e:
            logger.debug(str(e))
            self.error = "Unable to load JSON data"
            self.errorCode = -1
        
    def _getBeepPath(self, video_rate):
        logger.debug('checking video_rate: {}'.format(video_rate))
        vids = {'23.976':"BEEPS_23976.wav", '24':"BEEPS_24.wav", '25':"BEEPS_25.wav", '30':"BEEPS_24.wav"}
        wrkdir = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(wrkdir, vids.get(video_rate, "BEEPS_24.wav"))

    def _getTenByTenPath(self, video_rate):
        logger.debug('checking video_rate: {}'.format(video_rate))
        vids = {'23.976':"tenbyten2398.wav", '24':"tenbyten24.wav", '25':"tenbyten25.wav", '30':"tenbyten30.wav"}
        wrkdir = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(wrkdir, vids.get(video_rate, "BEEPS_24.wav"))

    def frameOffsetForHour(self, hour):
        return 24 * 60 * 60 * hour

    def _writeAAFBeeps(self):
        logger.debug('Begin write aaf beeps')

        self.beep_path = self._getBeepPath(self.video_rate)
        if not os.path.exists(self.beep_path):
            self.error = "No Beeps available"
            self.errorCode = -1
            logger.debug(self.error)
            return

        self.beepsFolder = ''
        if self.shouldLink == True:
            """
            CREATE A FOLDER NEXT TO THE .aaf FILE,
            AND UPDATE THE BEEPS PATH AS WE'LL NEED AT LEAST ONE FILE
            """
            logger.debug('Will link media and use symbolic links')
            name = os.path.splitext(self.destinationFile)[0]
            name += "_beeps"
            self.beepsFolder = os.path.join(os.path.dirname(self.destinationFile), name)
            try:
                if os.path.exists(self.beepsFolder):
                    logger.debug('Removing old beep folder: {}'.format(self.beepsFolder))
                    shutil.rmtree(self.beepsFolder)
                os.mkdir(self.beepsFolder)
                updatedBeepPath = os.path.join(self.beepsFolder, os.path.basename(self.beep_path))
                shutil.copyfile(self.beep_path, updatedBeepPath)
                self.beep_path = updatedBeepPath
            except Exception as e:
                self.error = 'Unable to create folder for beeps, the folder might already exist:{}'.format(str(e))
                self.errorCode = -1
                logger.debug(self.error)
                return
            
        else:
            """
            CREATE STAGEING LOCATION FOR UNIQUE ESSENCE
            """
            logger.debug('Creating embedded media')
            self.beepsFolder = mkdtemp(prefix="ADT__")
            logger.debug('using the following beep folder {}'.format(self.beepsFolder))
            
        with aaf2.open(self.destinationFile, "w") as f:
            logger.debug('Creating AAF file at {}'.format(self.destinationFile))

            comp_mob = f.create.CompositionMob("composition_mob")
            f.content.mobs.append(comp_mob)

            tc_segment = f.create.Timecode(int(float(self.video_rate)+0.5), drop=False)
            tc_segment.start = self.startInFrames
            tc_slot = comp_mob.create_empty_sequence_slot(self.video_rate, slot_id=len(self.tracks)+1, media_kind="timecode")
            tc_slot.name = "Timecode"
            tc_slot.segment = tc_segment

            BEEP_LENGTH = 48
            for track_number, cues in enumerate(self.tracks):
                logger.debug('iterating over tracks') 

                # This sequence will be added to the composition mob
                sequence = f.create.Sequence(media_kind="sound")
                # sequences.append(sequence)

                currentFrame = 0
                lastLength = 0

                for i, cue in enumerate(cues):
                    cue['startTime'] -= BEEP_LENGTH
                    cue['endTime'] -= BEEP_LENGTH
 
                current_pos = 0
                for i, cue in enumerate(cues):
                    # this is the length of the clip on the timeline in frames
                    pad = cue['startTime'] - current_pos

                    # Set the origin of the source_mob's first slot to set an internal offset
                    nextBeepPath = os.path.join(self.beepsFolder, "BEEPS_{}.wav".format(str(uuid4())))
                    # os.symlink(self.beep_path, nextBeepPath)
                    shutil.copy(self.beep_path, nextBeepPath)

                    if self.shouldLink == True:
                        info = meta(self.video_rate)
                        info['format']['filename'] = nextBeepPath
                        master_mob, source_mob, tape_mob = f.content.link_external_wav(info)
                        # Set the origin of the underlying clip,
                        # This would be the offset into the file
                        # source_mob.slots[0].origin = 0;
                    else:
                        # now finally import the generated media
                        master_mob = f.create.MasterMob("Beeps")
                        f.content.mobs.append(master_mob)
                        master_mob.import_audio_essence(nextBeepPath, self.video_rate)

                    # Set the name of the clip
                    # This would be the cue name, whose length I should check in Protools
                    master_mob.name = cue['name']

                    comp_fill = f.create.Filler("sound", pad)
                    master_clip = master_mob.create_source_clip(slot_id=1, length=BEEP_LENGTH) #cue['endTime'] - cue['startTime']
                    sequence.components.append(comp_fill)
                    sequence.components.append(master_clip)
                    current_pos += (pad + BEEP_LENGTH)
            
                timeline_slot = comp_mob.create_timeline_slot(self.video_rate)
                timeline_slot.name = "Track {}".format(track_number+1)
                timeline_slot.segment= sequence
                timeline_slot.slot_id= track_number+1001

    def _writeAAFTenByTen(self):
        logger.debug('Begin write 10x10s')

        self.beep_path = self._getTenByTenPath(self.video_rate)
        if not os.path.exists(self.beep_path):
            self.error = "No 10x10s available"
            self.errorCode = -1
            logger.debug(self.error)
            return

        self.beepsFolder = ''
        if self.shouldLink == True:
            """
            CREATE A FOLDER NEXT TO THE .aaf FILE,
            AND UPDATE THE BEEPS PATH AS WE'LL NEED AT LEAST ONE FILE
            """
            logger.debug('Will link media and use symbolic links')
            name = os.path.splitext(self.destinationFile)[0]
            name += "_beeps"
            self.beepsFolder = os.path.join(os.path.dirname(self.destinationFile), name)
            try:
                if os.path.exists(self.beepsFolder):
                    logger.debug('Removing old beep folder: {}'.format(self.beepsFolder))
                    shutil.rmtree(self.beepsFolder)
                os.mkdir(self.beepsFolder)
                updatedBeepPath = os.path.join(self.beepsFolder, os.path.basename(self.beep_path))
                shutil.copyfile(self.beep_path, updatedBeepPath)
                self.beep_path = updatedBeepPath
            except Exception as e:
                self.error = 'Unable to create folder for beeps, the folder might already exist:{}'.format(str(e))
                self.errorCode = -1
                logger.debug(self.error)
                return
            
        else:
            """
            CREATE STAGEING LOCATION FOR UNIQUE ESSENCE
            """
            logger.debug('Creating embedded media')
            self.beepsFolder = mkdtemp(prefix="ADT__")
            logger.debug('using the following beep folder {}'.format(self.beepsFolder))
            
        with aaf2.open(self.destinationFile, "w") as f:
            logger.debug('Creating AAF file at {}'.format(self.destinationFile))

            comp_mob = f.create.CompositionMob("composition_mob")
            f.content.mobs.append(comp_mob)

            tc_segment = f.create.Timecode(int(float(self.video_rate)+0.5), drop=False)
            tc_segment.start = self.startInFrames
            tc_slot = comp_mob.create_empty_sequence_slot(self.video_rate, slot_id=len(self.tracks)+1, media_kind="timecode")
            tc_slot.name = "Timecode"
            tc_slot.segment = tc_segment

            TEN_BY_TEN_LENGTH = 20
            for track_number, cues in enumerate(self.tracks):
                logger.debug('iterating over tracks') 

                # This sequence will be added to the composition mob
                sequence = f.create.Sequence(media_kind="sound")
                # sequences.append(sequence)

                currentFrame = 0
                lastLength = 0

                for i, cue in enumerate(cues):
                    # print cue['startTime'],  cue['endTime']
                    cue['startTime'] -= TEN_BY_TEN_LENGTH
                    cue['endTime'] -= TEN_BY_TEN_LENGTH
                    # print cue['startTime'],  cue['endTime']
 
                current_pos = 0
                for i, cue in enumerate(cues):
                    # this is the length of the clip on the timeline in frames
                    pad = cue['startTime'] - current_pos

                    # Set the origin of the source_mob's first slot to set an internal offset
                    nextBeepPath = os.path.join(self.beepsFolder, "BEEPS_{}.wav".format(str(uuid4())))
                    # os.symlink(self.beep_path, nextBeepPath)
                    shutil.copy(self.beep_path, nextBeepPath)

                    if self.shouldLink == True:
                        info = meta(self.video_rate)
                        info['format']['filename'] = nextBeepPath
                        master_mob, source_mob, tape_mob = f.content.link_external_wav(info)
                        # Set the origin of the underlying clip,
                        # This would be the offset into the file
                        # source_mob.slots[0].origin = 0;
                    else:
                        # now finally import the generated media
                        master_mob = f.create.MasterMob("Beeps")
                        f.content.mobs.append(master_mob)
                        master_mob.import_audio_essence(nextBeepPath, self.video_rate)

                    # Set the name of the clip
                    # This would be the cue name, whose length I should check in Protools
                    master_mob.name = 'BEEPS' #cue['name']

                    comp_fill = f.create.Filler("sound", pad)
                    master_clip = master_mob.create_source_clip(slot_id=1, length=TEN_BY_TEN_LENGTH) #cue['endTime'] - cue['startTime']
                    sequence.components.append(comp_fill)
                    sequence.components.append(master_clip)
                    current_pos += (pad + TEN_BY_TEN_LENGTH)
            
                timeline_slot = comp_mob.create_timeline_slot(self.video_rate)
                timeline_slot.name = "Track {}".format(track_number+1)
                timeline_slot.segment= sequence
                timeline_slot.slot_id= track_number+1001


    def _writeAAFClips(self):
        logger.debug('Begin write aaf clips')

        self.clipsFolder = ''
        if self.shouldLink == True:
            """
            DONT THINK I AM USING THIS AT PRESENT
            CREATE A FOLDER NEXT TO THE .aaf FILE,
            AND UPDATE THE BEEPS PATH AS WE'LL NEED AT LEAST ONE FILE
            """
            logger.debug('Will link media and use symbolic links')
            name = os.path.splitext(self.destinationFile)[0]
            name += "_clips"
            self.clipsFolder = os.path.join(os.path.dirname(self.destinationFile), name)
            try:
                if os.path.exists(self.clipsFolder):
                    logger.debug('Removing old clips folder: {}'.format(self.clipsFolder))
                    shutil.rmtree(self.clipsFolder)
                os.mkdir(self.clipsFolder)
                updatedBeepPath = os.path.join(self.clipsFolder, os.path.basename(self.beep_path))
                shutil.copyfile(self.beep_path, updatedBeepPath)
                self.beep_path = updatedBeepPath
            except Exception as e:
                self.error = 'Unable to create folder for the clips, the folder might already exist:{}'.format(str(e))
                self.errorCode = -1
                logger.debug(self.error)
                return
            
        else:
            """
            CREATE STAGEING LOCATION FOR UNIQUE ESSENCE
            """
            logger.debug('Creating embedded media')
            self.clipsFolder = mkdtemp(prefix="ADT__")
            logger.debug('using the following clips folder {}'.format(self.clipsFolder))
            
        with aaf2.open(self.destinationFile, "w") as f:
            logger.debug('Creating AAF file at {}'.format(self.destinationFile))

            comp_mob = f.create.CompositionMob("composition_mob")
            f.content.mobs.append(comp_mob)

            tc_segment = f.create.Timecode(int(float(self.video_rate)+0.5), drop=False)
            tc_segment.start = self.startInFrames
            tc_slot = comp_mob.create_empty_sequence_slot(self.video_rate, slot_id=len(self.tracks)+1, media_kind="timecode")
            tc_slot.name = "Timecode"
            tc_slot.segment = tc_segment

            for track_number, cues in enumerate(self.tracks):
                logger.debug('iterating over tracks') 

                # This sequence will be added to the composition mob
                sequence = f.create.Sequence(media_kind="sound")
                # sequences.append(sequence)

                currentFrame = 0
                lastLength = 0
                current_pos = 0
                for i, cue in enumerate(cues):
                    # this is the length of the clip on the timeline in frames
                    pad = cue['startTime'] - current_pos
                    cueLength = cue['endTime'] - cue['startTime']
                    if cueLength < 10:
                        cueLength == 10
                    spf = meta(self.video_rate)['samples_per_frame']

                    # Set the origin of the source_mob's first slot to set an internal offset
                    nextClipPath = os.path.join(self.clipsFolder, "CLIP_{}.wav".format(str(uuid4())))
                    # os.symlink(self.beep_path, nextClipPath)

                    #TODO MAKE AUDIO SILENCE WITH LENGTH 
                    # nextSilenceFile = GEN_SILENCE_OF_LENGTH(DIFF_IN_FRAMES*SAMPLES_PER_FRAME)
                    subprocess.call(['./createaudiosilence', nextClipPath, '24', '48000', '{}'.format(spf*cueLength)])

                    if self.shouldLink == True:
                        info = meta(self.video_rate)
                        info['format']['filename'] = nextClipPath
                        master_mob, source_mob, tape_mob = f.content.link_external_wav(info)
                        # Set the origin of the underlying clip,
                        # This would be the offset into the file
                        # source_mob.slots[0].origin = 0;
                    else:
                        # now finally import the generated media
                        master_mob = f.create.MasterMob("Clips")
                        f.content.mobs.append(master_mob)
                        master_mob.import_audio_essence(nextClipPath, self.video_rate)

                    # Set the name of the clip
                    # This would be the cue name, whose length I should check in Protools
                    try:
                        line = cue['line']
                        lnlength = 15
                        line = (cue['line'][:lnlength] + '...') if len(cue['line']) > lnlength else cue['line']
                        # master_mob.name = cue['name'] + "_" + line
                        master_mob.name = cue['name']
                    except Exception as e:
                        master_mob.name = cue['name']

                    comp_fill = f.create.Filler("sound", pad)
                    master_clip = master_mob.create_source_clip(slot_id=1, length=cueLength)
                    sequence.components.append(comp_fill)
                    sequence.components.append(master_clip)
                    current_pos += (pad + cueLength)
            
                timeline_slot = comp_mob.create_timeline_slot(self.video_rate)
                timeline_slot.name = "Track {}".format(track_number+1)
                timeline_slot.segment= sequence
                timeline_slot.slot_id= track_number+1001

if __name__ == '__main__':
    logger = createLoggerWithPath2(os.path.expanduser('~/Library/Application Support/ADRToolBox/aaflog.txt'))
    logger.debug('Starting Beep Creation')
    logger.debug(sys.argv[1])
    logger.debug(sys.argv[2])
    logger.debug(sys.argv[3])
    AAFWriter(sys.argv[1], sys.argv[2], sys.argv[3])
    logger.debug('Finished')
